from numbers import Number

import traits.api as tr
from bmcs_beam.beam_config.system.cantilever_system import CantileverDistLoadSystem
from bmcs_beam.beam_config.system.four_pb_system import FourPBSystem
from bmcs_beam.beam_config.system.simple_dist_load_system import SimpleDistLoadSystem
from bmcs_beam.beam_config.system.three_pb_system import ThreePBSystem
from bmcs_cross_section.cs_design import CrossSectionDesign
from bmcs_utils.api import Item, View, EitherType, Str, Bool


class BeamDesign(CrossSectionDesign):

    name = 'BeamSystem Design'

    system = EitherType(options=[('3pb', ThreePBSystem),
                                 ('4pb', FourPBSystem),
                                 ('simple_beam_dist_load', SimpleDistLoadSystem),
                                 ('cantilever_dist_load', CantileverDistLoadSystem),
                                 # ('fixed_support_dist_load', CarbonReinfMatMod),
                                 # ('fixed_and_roller_support_dist_load', CarbonReinfMatMod),
                                 # ('three_span_dist_load', CarbonReinfMatMod),
                                 # ('three_pb_fixed_support', CarbonReinfMatMod),
                                 # ('single_moment', CarbonReinfMatMod),
                                 ])

    depends_on = ['matrix', 'cross_section_layout', 'cross_section_shape', 'system']
    tree = ['system']

    L = tr.Property(depends_on='state_changed')
    @tr.cached_property
    def _get_L(self):
        return self.system_.L
    def _set_L(self, L):
        self.system_.L = L

    # F = tr.DelegatesTo('system_')
    # x = tr.DelegatesTo('system_')

    ipw_view = View(
        Item('system'),
    )

    def get_M_x(self):
        return self.system_.get_moment()

    def get_Q_x(self):
        return self.system_.get_shear()

    def subplots(self, fig):
        return self.system_.subplots(fig)

    # @todo: [HS] this limits the beam design object to just a MQ profile
    # it is, however meant to be only useful for the deflection calculation.
    # For the shear zone - the interface should be generalized and the
    # beam design should provide more services, namely the access to
    # all the components of the beam, draw the plan from the side view,
    # cross sectional view - indeed the whole design including the supports
    # and loading.
    # **[HS]: I would take this second variant**
    # An alternative is to declare this package just to a beam_statical_system
    # which is purely 1D and not to call it design. Then a BeamDesign would
    # contain both - beam_bc (a statical beam) and cs_design. It would provide
    # functionality to change the parameters of the design - including the
    # length, load configuration and supports (BC in one word).
    #
    def update_plot(self, axes):
        print('plot updated!')
        self.system_.update_plot(axes)

    # Quick fix: needed for [bmcs_shear_zone]
    def plot_reinforcement(self, ax):
        L = self.system_.L
        for z in self.cross_section_layout.reinforcement.z_j:
            ax.plot([0,L],[z,z], lw=2, color='brown')


'''
Created on 12.01.2016
@author: RChudoba, ABaktheer, Yingxiong

@todo: derive the size of the state array.
'''

from ibvpy.tfunction import MonotonicLoadingScenario, CyclicLoadingScenario
from ibvpy.api import \
    BCSlice, Hist
from ibvpy.bcond.bcond_mngr import BCondMngr
from ibvpy.fets import \
    FETS2D4Q
from ibvpy.tmodel.mats2D import \
    MATS2DElastic, MATS2DMplDamageEEQ, MATS2DScalarDamage, \
    MATS2DMplCSDEEQ
from scipy import interpolate as ip
from ibvpy.api import TStepBC, XDomainFEGrid
from traits.api import \
    Property, cached_property
from bmcs_utils.api import \
    Float, Int, Instance, View, Item, EitherType, Model, ProgressEditor, \
    HistoryEditor
from matplotlib import cm
from ibvpy.view.plot2d import Viz2D, Vis2D
from ibvpy.view.ui import BMCSLeafNode

import numpy as np
import traits.api as tr

from matplotlib.patches import PathPatch
from matplotlib.path import Path
import matplotlib.patches as mpatches
from .viz3d_energy import ForceDisplacement, Vis2DEnergy, Vis2DCrackBand

class Viz2DForceDeflection(Viz2D):
    '''Plot adaptor for the pull-out simulator.
    '''
    label = 'F-w'

    show_legend = tr.Bool(True, auto_set=False, enter_set=True)

    def plot(self, ax, vot, *args, **kw):
        sim = self.vis2d.sim
        P, W = sim.hist['Pw'].Pw
        ymin, ymax = np.min(P), np.max(P)
        L_y = ymax - ymin
        ymax += 0.05 * L_y
        ymin -= 0.05 * L_y
        xmin, xmax = np.min(W), np.max(W)
        L_x = xmax - xmin
        xmax += 0.03 * L_x
        xmin -= 0.03 * L_x
        color = kw.get('color', 'black')
        linewidth = kw.get('linewidth', 2)
        label = kw.get('label', 'P(w)')
        ax.plot(W, P, linewidth=linewidth, color=color, alpha=0.4,
                label=label)
        ax.set_ylim(ymin=ymin, ymax=ymax)
        ax.set_xlim(xmin=xmin, xmax=xmax)
        ax.set_ylabel('Force P [N]')
        ax.set_xlabel('Deflection w [mm]')
        if self.show_legend:
            ax.legend(loc=4)
        self.plot_marker(ax, vot)

    def plot_marker(self, ax, vot):
        sim = self.vis2d.sim
        P, W = sim.hist['Pw'].Pw
        idx = sim.hist.get_time_idx(vot)
        P, w = P[idx], W[idx]
        ax.plot([w], [P], 'o', color='black', markersize=10)

    def plot_tex(self, ax, vot, *args, **kw):
        self.plot(ax, vot, *args, **kw)

    traits_view = View(
        Item('name', style='readonly'),
        Item('show_legend'),
    )

def align_xaxis(ax1, ax2):
    """Align zeros of the two axes, zooming them out by same ratio"""
    axes = np.array([ax1, ax2])
    (ax1_min, ax1_max), (ax2_min, ax2_max) = np.array([ax.get_xlim() for ax in axes])
    ax1_delta = ax1_max - ax1_min
    ax2_delta = ax2_max - ax2_min
    ax1_rmin, ax1_rmax = ax1_min / ax1_delta, ax1_max / ax1_delta
    ax2_rmin, ax2_rmax = ax2_min / ax2_delta, ax2_max / ax2_delta
    rmin = np.min([ax1_rmin, ax2_rmin])
    rmax = np.max([ax1_rmax, ax2_rmax])
    ax1_min, ax1_max = ax1_delta * rmin, ax1_delta * rmax
    ax2_min, ax2_max = ax2_delta * rmin, ax2_delta * rmax
    ax1.set_xlim(ax1_min, ax1_max)
    ax2.set_xlim(ax2_min, ax2_max)


class Viz2DStrainInCrack(Viz2D):

    '''Plot adaptor for the pull-out simulator.
    '''
    label = 'strain in crack'

    show_legend = tr.Bool(True, auto_set=False, enter_set=True)

    ax_sig = tr.Any

    def plot(self, ax, vot, *args, **kw):
        eps = self.vis2d.get_eps_t(vot)
        sig = self.vis2d.get_sig_t(vot)
        a_x = self.vis2d.get_a_x()
        ax.plot(eps, a_x, linewidth=3, color='red', alpha=0.4,
                label='P(w;x=L)')
#        ax.fill_betweenx(eps, 0, a_x, facecolor='orange', alpha=0.2)
        ax.set_xlabel('Strain [-]')
        ax.set_ylabel('Position [mm]')
        if self.ax_sig:
            self.ax_sig.clear()
        else:
            self.ax_sig = ax.twiny()
        self.ax_sig.set_xlabel('Stress [MPa]')
        self.ax_sig.plot(sig, a_x, linewidth=2, color='blue')
        self.ax_sig.fill_betweenx(a_x, sig, facecolor='blue', alpha=0.2)
        align_xaxis(ax, self.ax_sig)
        if self.show_legend:
            ax.legend(loc=4)

    traits_view = View(
        Item('name', style='readonly'),
        Item('show_legend'),
    )


class Viz2DStressInCrack(Viz2D):

    '''Plot adaptor for the pull-out simulator.
    '''
    label = 'stress in crack'

    show_legend = tr.Bool(True, auto_set=False, enter_set=True)

    def plot(self, ax, vot, *args, **kw):
        sig = self.vis2d.get_sig_t(vot)
        a_x = self.vis2d.get_a_x()
        ax.plot(a_x, sig, linewidth=3, color='red', alpha=0.4,
                label='P(w;x=L)')
        ax.fill_between(a_x, 0, sig, facecolor='orange', alpha=0.2)
        ax.set_ylabel('Stress [-]')
        ax.set_xlabel('Position [mm]')
        if self.show_legend:
            ax.legend(loc=4)

    traits_view = View(
        Item('name', style='readonly'),
        Item('show_legend'),
    )


class Viz2DTA(Viz2D):

    '''Plot adaptor for unstable tensile cracking.
    '''
    label = 'crack length'

    show_legend = tr.Bool(True, auto_set=False, enter_set=True)

    def plot(self, ax, vot, *args, **kw):
        t = self.vis2d.get_t()
        a = self.vis2d.get_a_t()
        ax.plot(t, a, linewidth=3, color='blue', alpha=0.4,
                label='Crack length')
        ax.fill_between(t, 0, a, facecolor='blue', alpha=0.2)
        ax.set_xlabel('time')
        ax.set_ylabel('a')
        if self.show_legend:
            ax.legend(loc=4)

    traits_view = View(
        Item('name', style='readonly'),
        Item('show_legend'),
    )


class Viz2DdGdA(Viz2D):
    label = 'Energy release per unit crack length'

    show_legend = tr.Bool(True, auto_set=False, enter_set=True)

    vis2d_cb = tr.WeakRef

    def plot(self, ax, vot, *args, **kw):
        t = self.vis2d.sim.hist.t
        G_t = self.vis2d.get_G_t()
        a_t = self.vis2d_cb.get_a_t()
        b = self.vis2d_cb.sim.cross_section.B

        tck = ip.splrep(a_t * b, G_t, s=0, k=1)
        dG_da = ip.splev(a_t, tck, der=1)

#         ax.plot(a_t, dG_da, linewidth=3, color='blue', alpha=0.4,
#                 label='dG/da')
#         ax.fill_between(a_t, 0, dG_da, facecolor='blue', alpha=0.2)
#         ax.set_xlabel('time')
#         ax.set_ylabel('dG_da')

        tck = ip.splrep(t, G_t, s=0, k=1)
        dG_dt = ip.splev(t[:-1], tck, der=1)
        tck = ip.splrep(t, a_t, s=0, k=1)
        da_dt = ip.splev(t[:-1], tck, der=1)
        nz = da_dt != 0.0
        dG_da = np.zeros_like(da_dt)
        dG_da[nz] = dG_dt[nz] / da_dt[nz] / b
        ax.plot(a_t[1:], dG_da, linewidth=3, color='green', alpha=0.4,
                label='dG/dt / da_dt')

#         ax.plot(a_t, G_t, linewidth=3, color='black')
#         ax.fill_between(a_t, 0, G_t, facecolor='blue', alpha=0.2)

        if self.show_legend:
            ax.legend(loc=4)

    traits_view = View(
        Item('name', style='readonly'),
        Item('show_legend'),
    )

class RestrainedCrackingHistory(Hist, Model):

    vis_record = tr.Dict
    def _vis_record_default(self):
        return {'Fw': ForceDisplacement(),
                'energy' : Vis2DEnergy(),
                'crack_band' : Vis2DCrackBand() }

    t_slider = Float(0)
    t_max = tr.Property()
    def _get_t_max(self):
        return self.t[-1]

    def plot_Fw(self, ax):
        F_range, w_range = self['Fw'].Fw
#        F_range, w_range = self.get_Fw_t()
        idx = self.get_time_idx(self.t_slider)
        F = F_range[idx]
        w = w_range[idx]
        ax.plot(w, F * 0.001, marker='o', color='magenta')
        ax.plot(w_range, F_range * 0.001)
        ax.set_ylabel(r'$F$ [kN]')
        ax.set_xlabel(r'$w$ [mm]')
        # ax.legend()

    def plot_sig_eps(self, ax):
        L = self.tstep_source.geometry.L
        H = self.tstep_source.geometry.H
        B = self.tstep_source.cross_section.B
        A = H * B
        F_range, w_range = self['Fw'].Fw
        sig_range = F_range / A
        eps_range = w_range / L
        ax.plot(eps_range, sig_range)
        ax.set_ylabel(r'$\sigma$ [MPa]')
        ax.set_xlabel(r'$\varepsilon$ [-]')

    U_bar_t = tr.Property
    def _get_U_bar_t(self):
        vis_energy = self['energy']
        return vis_energy.U_bar_t

    W_t = tr.Property
    def _get_W_t(self):
        vis_energy = self['energy']
        return vis_energy.get_W_t()

    def plot_G_t(self, ax,
             label_U='U(t)', label_W='W(t)',
             color_U='green', color_W='black'):
        vis_energy = self['energy']
        t = vis_energy.get_t()
        U_bar_t = np.array(vis_energy.U_bar_t, dtype=np.float_)
        W_t = vis_energy.get_W_t()
        ax.plot(t, W_t, color=color_W, label=label_W)
        ax.plot(t, U_bar_t, color=color_U, label=label_U)
        # Energy contribution relevant evaluated using internal variables
        # G_omega_t = np.array(vis_energy.G_omega_t, dtype=np.float_)
        # ax.plot(t, U_bar_t+G_omega_t, color='black', linestyle='dashed', label='G_t')
        ax.fill_between(t, W_t, U_bar_t, facecolor='gray', alpha=0.5,
                        label='G(t)')
        ax.fill_between(t, U_bar_t, 0, facecolor=color_U, alpha=0.5,
                        label='U(t)')
        ax.set_ylabel('energy [Nmm]')
        ax.set_xlabel('control displacement [mm]')
        ax.legend()

    def plot_crack_band(self, ax_eps, ax_sig):
        vis_cb = self['crack_band']
        t_idx = self.get_time_idx(self.t_slider)
        eps = vis_cb.get_eps_t(t_idx)
        sig = vis_cb.get_sig_t(t_idx)
        a_x = vis_cb.get_a_x()
        ax_eps.plot(eps, a_x, linewidth=1, color='red', alpha=0.4,
                label=r'$\varepsilon$ [-]')
        ax_eps.fill_betweenx(a_x, 0, eps, facecolor='orange', alpha=0.2)
        ax_eps.set_xlabel(r'$\varepsilon$ [-]')
        ax_eps.set_ylabel(r'$y$ [mm]')
        ax_sig.set_xlabel(r'$\sigma$ [MPa]')
        ax_sig.plot(sig, a_x, linewidth=1, color='blue',
                    label=r'$\sigma$ [MPa]')
        ax_sig.fill_betweenx(a_x, 0, sig, facecolor='blue', alpha=0.2)
        align_xaxis(ax_eps, ax_sig)
        # with bmcs_utils.api.print_output:
        #     print('eps', eps)
        ax_eps.legend(loc=5)
        ax_sig.legend(loc=4)

    warp_factor = Float(1.0)

    def plot_mesh(self, ax):
        tstep = self.tstep_source
        xdomain = tstep.xdomain
        t_idx = self.get_time_idx(self.t_slider)
        U = self.U_t[t_idx]
        o_Ia = xdomain.o_Ia  # bt.n_a + 1:]
        U_Ia = U[o_Ia] * self.warp_factor

        X_Ia = xdomain.mesh.X_Id + U_Ia # bt.n_a + 1:]
        I_Ei = xdomain.mesh.I_Ei
        i_lj = np.array([[0,1],[1,2],[2,3],[3,0]])
        X_Eia = X_Ia[I_Ei]
        X_Elja = X_Eia[:,i_lj,:]
        X_ajL = np.einsum('Elja->ajlE', X_Elja).reshape(2,2,-1)
        tstep.plot_geo(ax)
        ax.plot(X_ajL[0,...], X_ajL[1,...], color='black', linewidth=0.2)
        ax.axis('equal');

        omega_Em_t = np.array([state[0]['omega'] for state in self.state_vars])
        omega_Em = omega_Em_t[t_idx]
        n_Ii = np.bincount(I_Ei.flatten())
        omega_I = np.bincount(I_Ei.flatten(), weights=omega_Em.flatten()) / n_Ii
        X_aI = X_Ia.T
        n_x, n_y = tstep.n_e_x + 1, tstep.n_e_y + 1
        x_I, y_I = X_aI.reshape(2, n_x, n_y)
        omega_II = omega_I.reshape(n_x, n_y)
        ax.pcolormesh(x_I, y_I, omega_II, shading='nearest', cmap=cm.PuBu)
        ax.axis('equal')

    def subplots(self, fig):
        gs = fig.add_gridspec(2,2, width_ratios=[1., 1.])
        ax1 = fig.add_subplot(gs[0,0])
        ax2 = fig.add_subplot(gs[0,1])
        ax3 = fig.add_subplot(gs[1,0])
        ax4 = fig.add_subplot(gs[1,1])
        ax44 = ax4.twiny()
        return ax1, ax2, ax3, ax4, ax44

    def update_plot(self, axes):
        ax_mesh, ax_Fw, ax_energy, ax_eps, ax_sig = axes
        self.plot_mesh(ax_mesh)
        #self.plot_Fw(ax_Fw)
        self.plot_sig_eps(ax_Fw)
        self.plot_G_t(ax_energy)
        self.plot_crack_band(ax_eps, ax_sig)

    ipw_view = View(
        Item('warp_factor'),
        Item('t_slider', readonly=True),
        time_editor=HistoryEditor(
            var='t_slider',
            max_var='t_max',
        )
    )


class CrossSection(BMCSLeafNode):
    '''Parameters of the pull-out cross section
    '''
    name = 'cross-section'

    B = Float(100.0,
              CS=True,
              label='thickness',
              auto_set=False, enter_set=True,
              desc='cross-section width [mm2]')

    ipw_view = View(
        Item('B'),
    )


class Geometry(BMCSLeafNode):

    name = 'geometry'
    H = Float(100.0,
              label='beam depth',
              GEO=True,
              auto_set=False, enter_set=True,
              desc='cross section height [mm2]')
    L = Float(600.0,
              label='beam length',
              GEO=True,
              auto_set=False, enter_set=True,
              desc='Length of the specimen')
    a = Float(20.0,
              GEO=True,
              label='notch depth',
              auto_set=False, enter_set=True,
              desc='Depth of the notch')
    L_cb = Float(4.0,
                GEO=True,
                label='crack band width',
                auto_set=False, enter_set=True,
                desc='Width of the crack band')

    ipw_view = View(
        Item('H'),
        Item('L'),
        Item('a'),
        Item('L_cb', latex=r'L_\mathrm{cb}'),
    )


itags_str = "state_changed"

class RestrainedCrackingTestModel(TStepBC):

    #=========================================================================
    # Tree node attributes
    #=========================================================================
    node_name = 'RestrainedCracking test simulation'
    name = 'RestrainedCracking'

    hist_type = RestrainedCrackingHistory

    history = tr.Property()
    @tr.cached_property
    def _get_history(self):
        return self.hist

    time_line = tr.Property()
    @tr.cached_property
    def _get_time_line(self):
        return self.sim.tline

    depends_on = ['cross_section', 'geometry',
            'material_model', 'loading_scenario', 'hist']
    tree = ['cross_section', 'geometry',
            'material_model', 'loading_scenario', 'hist']

    # =========================================================================
    # Test setup parameters
    # =========================================================================
    loading_scenario = EitherType(
        options=[('monotonic', MonotonicLoadingScenario),
                 ('cyclic', CyclicLoadingScenario)],
        report=True, TIME=True,
        desc='object defining the loading scenario'
    )

    cross_section = Instance(CrossSection)

    def _cross_section_default(self):
        return CrossSection()

    geometry = Instance(Geometry)

    def _geometry_default(self):
        return Geometry()

    #=========================================================================
    # Discretization
    #=========================================================================
    n_e_x = Int(20,
                label='# of elems in x-dir',
                MESH=True, auto_set=False, enter_set=True)
    n_e_y = Int(8,
                label='# of elems in y-dir',
                MESH=True, auto_set=False, enter_set=True)

    w_max = Float(-5, BC=True, auto_set=False, enter_set=True)

    k_max = tr.Property(Int, depends_on='state_change')
    @tr.cached_property
    def _get_k_max(self):
        return self.sim.tloop.k_max
    def _set_k_max(self, value):
        self.sim.tloop.k_max = value

    controlled_elem = Property

    def _get_controlled_elem(self):
        return 0

    #=========================================================================
    # Material model
    #=========================================================================
    material_model = EitherType(
        options=[('scalar damage', MATS2DScalarDamage),
                 ('elastic', MATS2DElastic),
                 ('microplane damage (eeq)', MATS2DMplDamageEEQ),
                 ('microplane CSD (eeq)', MATS2DMplCSDEEQ),
                 ],
        MAT=True
    )
    '''Material model'''

    mm = Property

    def _get_mm(self):
        return self.material_model_

    material = Property

    def _get_material(self):
        return self.material_model_


    #=========================================================================
    # Simulator interface
    #=========================================================================
    def run(self):
        self.sim.run()

    def reset(self):
        self.sim.reset()

    t = tr.Property()

    def _get_t(self):
        return self.sim.t

    def _set_t(self, value):
        self.sim.t = value

    t_max = tr.Property()

    def _get_t_max(self):
        return self.sim.t_max

    def _set_t_max(self, value):
        self.sim.t_max = value

    interrupt = tr.Property()

    def _get_interrupt(self):
        return self.sim.interrupt

    def _set_interrupt(self, value):
        self.sim.interrupt = value

    #=========================================================================
    # Finite element type
    #=========================================================================
    fets_eval = Property(Instance(FETS2D4Q),
                         depends_on=itags_str)
    '''Finite element time stepper implementing the corrector
    predictor operators at the element level'''
    @cached_property
    def _get_fets_eval(self):
        return FETS2D4Q()

    bc = Property(Instance(BCondMngr),
                  depends_on=itags_str)
    '''Boundary condition manager
    '''
    @cached_property
    def _get_bc(self):
        return [self.fixed_y,
                self.fixed_x,
                self.control_bc]

    fixed_y = Property(depends_on=itags_str)
    '''Foxed boundary condition'''
    @cached_property
    def _get_fixed_y(self):
        return BCSlice(slice=self.fe_grid[-1, -1, -1, -1],
                       var='u', dims=[1], value=0)

    n_a = Property
    '''Element at the notch
    '''

    def _get_n_a(self):
        a_L = self.geometry.a / self.geometry.H
        return int(a_L * self.n_e_y)

    fixed_x = Property(depends_on=itags_str)
    '''Foxed boundary condition'''
    @cached_property
    def _get_fixed_x(self):
        # return BCSlice(slice=self.fe_grid[0, :, 0, :],
        #                var='u', dims=[0], value=0)
        return BCSlice(slice=self.fe_grid[0, self.n_a:, 0, :],
                       var='u', dims=[0], value=0)

    control_bc = Property(depends_on=itags_str)
    '''Control boundary condition - make it accessible directly
    for the visualization adapter as property
    '''
    @cached_property
    def _get_control_bc(self):
        return BCSlice(slice=self.fe_grid[-1, :, -1, :],
                       var='u', dims=[0], value=self.w_max)

    xdomain = Property(depends_on=itags_str)
    '''Discretization object.
    '''
    @cached_property
    def _get_xdomain(self):
        dgrid = XDomainFEGrid(coord_max=(self.geometry.L / 2., self.geometry.H),
                              integ_factor=self.cross_section.B,
                              shape=(self.n_e_x, self.n_e_y),
                              fets=self.fets_eval)

        L = self.geometry.L / 2.0
        L_cb = self.geometry.L_cb
        x_x, x_y = dgrid.mesh.geo_grid.point_x_grid
        L_1 = x_x[1, 0]
        d_L = L_cb - L_1
        x_x[1:, :] += d_L * (L - x_x[1:, :]) / (L - L_1)
        return dgrid

    fe_grid = Property

    def _get_fe_grid(self):
        return self.xdomain.mesh

    domains = Property(depends_on=itags_str)

    @cached_property
    def _get_domains(self):
        return [(self.xdomain, self.material_model_)]

    ipw_view = View(
        Item('w_max'),
        Item('k_max'),
        Item('n_e_x'),
        Item('n_e_y'),
        Item('material_model'),
        Item('loading_scenario'),
        time_editor=ProgressEditor(run_method='run',
                                      reset_method='reset',
                                      interrupt_var='interrupt',
                                      time_var='t',
                                      time_max='t_max',
                                      )
    )

    def plot_geo(self, ax):
        L, H = self.geometry.L, self.geometry.H
        L2 = L/2
        ax.plot([-L2, L2, L2, -L2, -L2], [0, 0, H, H, 0], color='black', linewidth=0.5)
        ax.fill([-L2, L2, L2, -L2, -L2], [0, 0, H, H, 0], color='gray', alpha=0.3)
        ax.set_xlabel(r'$x$ [mm]'); ax.set_ylabel(r'$y$ [mm]')
        ax.axis('equal');


        support_width = 0.04 * L
        support_line_distance = 1.3 * support_width

        supports = [(-L2, 0), (L2, 0)]
        codes = []
        vertices = []
        for i, support in enumerate(supports):
            x_loc, def_value = support
            # Draw the triangle of the support
            codes += [Path.MOVETO, Path.LINETO, Path.LINETO, Path.CLOSEPOLY]
            vertices += [(x_loc, 0),
                         (x_loc + support_width / 2, -support_width),
                         (x_loc - support_width / 2, -support_width),
                         (x_loc, 0)]

        vertices = np.array(vertices, np.float_)
        path = Path(vertices, codes)
        path_patch = PathPatch(path, facecolor='gray', edgecolor='gray')
        ax.add_patch(path_patch)
        return
        x_tail = 0
        y_tail = H * 1.8
        x_head = 0
        y_head = H
        xy_annotate = (-L2*0.3, H*2)
        arrow = mpatches.FancyArrowPatch((x_tail, y_tail), (x_head, y_head),
                                         color='blue', mutation_scale=10)
        ax.annotate('$w_\max$ = {} mm'.format(self.w_max), xy=xy_annotate, color='black')
        ax.add_patch(arrow)

    def subplots(self, fig):
        (ax_geo, ax_Fw), (ax_energy, ax_eps) = fig.subplots(2, 2)
        self.plot_geo(ax_geo)
        ax_sig = ax_eps.twiny()
        return ax_geo, ax_Fw, ax_energy, ax_eps, ax_sig

    def update_plot(self, axes):
        if len(self.history.U_t) == 0:
            return
        ax_geo, ax_Fw, ax_energy, ax_eps, ax_sig = axes
        self.history.plot_mesh(ax_geo)
        self.history.t_slider = self.t
        self.history.plot_Fw(ax_Fw)
        self.history.plot_G_t(ax_energy)
        self.history.plot_crack_band(ax_eps, ax_sig)




from bmcs_utils.api import Model, Int, Item, View, Float
from bmcs_beam.beam_config.system.beam_system import BeamSystem
import numpy as np

class ThreePBSystem(BeamSystem):

    name = 'ThreePBSystem'

    F = Float(-1000, SYSTEM=True)
    n_x = 2

    tree = []

    ipw_view = View(
        *BeamSystem.ipw_view.content,
        Item('F', latex='F \mathrm{[N]}', readonly=True),
    )

    def update_struct(self):
        self.struct = self.get_new_struct()
        n_x_to_F = round(0.5 * self.n_x)
        n_x_remains = self.n_x - n_x_to_F
        if n_x_to_F == 1:
            self.struct.add_element([[0, 0], [self.L / 2, 0]])
        else:
            self.struct.add_multiple_elements([[0, 0], [self.L / 2, 0]], n_x_to_F)
        if n_x_remains == 1:
            self.struct.add_element([[self.L / 2, 0], [self.L, 0]])
        else:
            self.struct.add_multiple_elements([[self.L / 2, 0], [self.L, 0]], n_x_remains)

        self.struct.add_support_hinged(1)
        self.struct.add_support_roll(self.n_x + 1)

        self.struct.point_load(n_x_to_F + 1, Fy=self.F)
        self.struct.solve()

    def get_max_force(self, M_I):
        # This should be done numerically, but here for efficiency and
        # because this is a known case it's given directly
        return 4 * np.max(M_I) / self.L

    def get_plot_force_scale_and_unit(self):
        """ Scale which should be applied on the force when plotting """
        return 1/1000, 'kN'


from bmcs_beam.beam_config.beam_design import BeamDesign
from bmcs_beam.beam_config.boundary_conditions import BoundaryConditions, BoundaryConfig
from bmcs_beam.bending.deflection_profile import DeflectionProfile, LoadDeflectionParamsStudy
from bmcs_beam.bending.beam_sls_curve import BeamSLSCurve, SLSParamStudy
from bmcs_beam.bending.beam_sls_curve import BeamSLSCurve
from bmcs_beam.beam_config.system.three_pb_system import ThreePBSystem
from bmcs_beam.beam_config.system.four_pb_system import FourPBSystem
from bmcs_beam.beam_config.system.simple_dist_load_system import SimpleDistLoadSystem
from bmcs_beam.beam_config.system.cantilever_system import CantileverDistLoadSystem



from anastruct.fem.plotter.mpl import Plotter
import numpy as np
import math
import matplotlib.pyplot as plt  # type: ignore
import matplotlib.patches as mpatches  # type: ignore
from anastruct.basic import find_nearest, rotate_xy
from anastruct.fem.plotter.values import (
    PlottingValues,
    det_scaling_factor,
    plot_values_axial_force,
    plot_values_bending_moment,
    plot_values_deflection,
    plot_values_element,
    plot_values_shear_force,
)

PATCH_SIZE = 0.03

class CustomPlotter(Plotter):
    def __init__(self, system, mesh):
        super().__init__(system, mesh)

    def __start_plot(self, figsize):
        plt.close("all")
        self.fig = plt.figure(figsize=figsize)
        self.one_fig = self.fig.add_subplot(111)
        plt.tight_layout()

    def __fixed_support_patch(self, max_val, ax=None):
        """
        :param max_val: max scale of the plot
        """
        if ax is None:
            ax = self.one_fig
        width = height = PATCH_SIZE * max_val
        for node in self.system.supports_fixed:
            support_patch = mpatches.Rectangle(
                (node.vertex.x - width * 0.5, -node.vertex.z - width * 0.5),
                width,
                height,
                color="r",
                zorder=9,
            )
            ax.add_patch(support_patch)

    def __hinged_support_patch(self, max_val, ax=None):
        """
        :param max_val: max scale of the plot
        """
        if ax is None:
            ax = self.one_fig
        radius = PATCH_SIZE * max_val
        for node in self.system.supports_hinged:
            support_patch = mpatches.RegularPolygon(
                (node.vertex.x, node.vertex.y - radius),
                numVertices=3,
                radius=radius,
                color="r",
                zorder=9,
            )
            ax.add_patch(support_patch)

    def __rotational_support_patch(self, max_val, ax=None):
        """
        :param max_val: max scale of the plot
        """
        if ax is None:
            ax = self.one_fig
        width = height = PATCH_SIZE * max_val
        for node in self.system.supports_rotational:
            support_patch = mpatches.Rectangle(
                (node.vertex.x - width * 0.5, -node.vertex.z - width * 0.5),
                width,
                height,
                color="r",
                zorder=9,
                fill=False,
            )
            ax.add_patch(support_patch)

    def __roll_support_patch(self, max_val, ax=None):
        """
        :param max_val: max scale of the plot
        """
        if ax is None:
            ax = self.one_fig
        radius = PATCH_SIZE * max_val
        count = 0
        for node in self.system.supports_roll:
            direction = self.system.supports_roll_direction[count]
            rotate = self.system.supports_roll_rotate[count]
            x1 = np.cos(np.pi) * radius + node.vertex.x + radius  # vertex.x
            z1 = np.sin(np.pi) * radius + node.vertex.y  # vertex.y
            x2 = (
                np.cos(np.radians(90)) * radius + node.vertex.x + radius
            )  # vertex.x + radius
            z2 = np.sin(np.radians(90)) * radius + node.vertex.y  # vertex.y + radius
            x3 = (
                np.cos(np.radians(270)) * radius + node.vertex.x + radius
            )  # vertex.x + radius
            z3 = np.sin(np.radians(270)) * radius + node.vertex.y  # vertex.y - radius

            triangle = np.array([[x1, z1], [x2, z2], [x3, z3]])

            if node.id in self.system.inclined_roll:
                angle = self.system.inclined_roll[node.id]
                triangle = rotate_xy(triangle, angle + np.pi * 0.5)
                support_patch = plt.Polygon(triangle, color="r", zorder=9)
                ax.add_patch(support_patch)
                ax.plot(
                    triangle[1:, 0] - 0.5 * radius * np.sin(angle),
                    triangle[1:, 1] - 0.5 * radius * np.cos(angle),
                    color="r",
                )
                if not rotate:
                    rect_patch = mpatches.RegularPolygon(
                        (node.vertex.x, node - node.vertex.y),
                        numVertices=4,
                        radius=radius,
                        orientation=angle,
                        color="r",
                        zorder=9,
                        fill=False,
                    )
                    ax.add_patch(rect_patch)

            elif direction == 2:  # horizontal roll
                support_patch = mpatches.RegularPolygon(
                    (node.vertex.x, node.vertex.y - radius),
                    numVertices=3,
                    radius=radius,
                    color="r",
                    zorder=9,
                )
                ax.add_patch(support_patch)
                y = -node.vertex.z - 2 * radius
                ax.plot(
                    [node.vertex.x - radius, node.vertex.x + radius], [y, y], color="r"
                )
                if not rotate:
                    rect_patch = mpatches.Rectangle(
                        (node.vertex.x - radius / 2, -node.vertex.z - radius / 2),
                        radius,
                        radius,
                        color="r",
                        zorder=9,
                        fill=False,
                    )
                    ax.add_patch(rect_patch)
            elif direction == 1:  # vertical roll
                # translate the support to the node

                support_patch = mpatches.Polygon(triangle, color="r", zorder=9)
                ax.add_patch(support_patch)

                y = node.vertex.y - radius
                ax.plot(
                    [node.vertex.x + radius * 1.5, node.vertex.x + radius * 1.5],
                    [y, y + 2 * radius],
                    color="r",
                )
                if not rotate:
                    rect_patch = mpatches.Rectangle(
                        (node.vertex.x - radius / 2, -node.vertex.z - radius / 2),
                        radius,
                        radius,
                        color="r",
                        zorder=9,
                        fill=False,
                    )
                    ax.add_patch(rect_patch)
            count += 1

    def __rotating_spring_support_patch(self, max_val, ax=None):
        """
        :param max_val: max scale of the plot
        """
        if ax is None:
            ax = self.one_fig
        radius = PATCH_SIZE * max_val

        for node, _ in self.system.supports_spring_y:
            r = np.arange(0, radius, 0.001)
            theta = 25 * np.pi * r / (0.2 * max_val)
            x = np.cos(theta) * r + node.vertex.x
            y = np.sin(theta) * r - radius + node.vertex.y
            ax.plot(x, y, color="r", zorder=9)

            # Triangle
            support_patch = mpatches.RegularPolygon(
                (node.vertex.x, node.vertex.y - radius * 3),
                numVertices=3,
                radius=radius * 0.9,
                color="r",
                zorder=9,
            )
            ax.add_patch(support_patch)

    def __spring_support_patch(self, max_val, ax=None):
        """
        :param max_val: max scale of the plot
        """
        if ax is None:
            ax = self.one_fig
        h = PATCH_SIZE * max_val
        left = -0.5 * h
        right = 0.5 * h
        dh = 0.2 * h

        for node, _ in self.system.supports_spring_z:
            yval = np.arange(0, -9, -1) * dh + node.vertex.y
            xval = (
                np.array([0, 0, left, right, left, right, left, 0, 0]) + node.vertex.x
            )

            ax.plot(xval, yval, color="r", zorder=10)

            # Triangle
            support_patch = mpatches.RegularPolygon(
                (node.vertex.x, -node.vertex.z - h * 2.6),
                numVertices=3,
                radius=h * 0.9,
                color="r",
                zorder=10,
            )
            ax.add_patch(support_patch)

        for node, _ in self.system.supports_spring_x:
            xval = np.arange(0, 9, 1) * dh + node.vertex.x
            yval = (
                np.array([0, 0, left, right, left, right, left, 0, 0]) + node.vertex.y
            )

            ax.plot(xval, yval, color="r", zorder=10)

            # Triangle
            support_patch = mpatches.RegularPolygon(
                (node.vertex.x + h * 1.7, -node.vertex.z - h),
                numVertices=3,
                radius=h * 0.9,
                color="r",
                zorder=10,
            )
            ax.add_patch(support_patch)

    def __q_load_patch(self, max_val, verbosity, ax=None):
        """
        :param max_val: max scale of the plot

        xn1;yn1  q-load   xn1;yn1
        -------------------
        |__________________|
        x1;y1  element    x2;y2
        """
        if ax is None:
            ax = self.one_fig
        for q_id in self.system.loads_q.keys():
            el = self.system.element_map[q_id]
            if el.q_load[0] > 0:
                direction = 1
            else:
                direction = -1

            h1 = 0.05 * max_val * abs(el.q_load[0]) / self.max_q
            h2 = 0.05 * max_val * abs(el.q_load[1]) / self.max_q
            x1 = el.vertex_1.x
            y1 = el.vertex_1.y
            x2 = el.vertex_2.x
            y2 = el.vertex_2.y

            if el.q_direction == "y":
                ai = 0
            elif el.q_direction == "x":
                ai = 0.5 * np.pi
            else:
                ai = -el.angle

            # - value, because the positive z of the beam is opposite of positive y of the plotter
            xn1 = x1 + np.sin(ai) * h1 * direction
            yn1 = y1 + np.cos(ai) * h1 * direction
            xn2 = x2 + np.sin(ai) * h2 * direction
            yn2 = y2 + np.cos(ai) * h2 * direction
            qi = el.q_load[0]
            q = el.q_load[1]
            coordinates = ([x1, xn1, xn2, x2], [y1, yn1, yn2, y2])
            ax.plot(*coordinates, color="g")
            rec = plt.Polygon(np.vstack(coordinates).T, color="g", alpha=0.3)
            ax.add_patch(rec)

            ax.text(xn1, yn1, f"q={qi}", color="b", fontsize=9, zorder=10)
            ax.text(xn2, yn2, f"q={q}", color="b", fontsize=9, zorder=10)

            if verbosity == 0:
                # arrow
                pos = np.sqrt(((y1 - y2) ** 2) + ((x1 - x2) ** 2))
                cg = ((pos / 3) * (el.q_load[0] + 2 * el.q_load[1])) / (
                    el.q_load[0] + el.q_load[1]
                )
                height = math.sin(el.angle) * cg
                base = math.cos(el.angle) * cg

                len_x1 = np.sin(ai - np.pi) * 0.6 * h1 * direction
                len_x2 = np.sin(ai - np.pi) * 0.6 * h2 * direction
                len_y1 = np.cos(ai - np.pi) * 0.6 * h1 * direction
                len_y2 = np.cos(ai - np.pi) * 0.6 * h2 * direction
                step_x = np.linspace(xn1, xn2, 11)
                step_y = np.linspace(yn1, yn2, 11)
                step_len_x = np.linspace(len_x1, len_x2, 11)
                step_len_y = np.linspace(len_y1, len_y2, 11)
                average_h = (h1 + h2) / 2
                # fc = face color, ec = edge color

                # add multiple arrows to fill load
                for counter in range(len(step_x)):
                    if q + qi >= 0:
                        if counter == 0:
                            shape = "right"
                        elif counter == 10:
                            shape = "left"
                        else:
                            shape = "full"
                    else:
                        if counter == 0:
                            shape = "left"
                        elif counter == 10:
                            shape = "right"
                        else:
                            shape = "full"

                    ax.arrow(
                        step_x[counter],
                        step_y[counter],
                        step_len_x[counter],
                        step_len_y[counter],
                        head_width=average_h * 0.25,
                        head_length=0.4
                        * np.sqrt(step_len_y[counter] ** 2 + step_len_x[counter] ** 2),
                        ec="k",
                        fc="k",
                        shape=shape,
                    )

    @staticmethod
    def __arrow_patch_values(Fx, Fz, node, h):
        """
        :param Fx: (float)
        :param Fz: (float)
        :param node: (Node object)
        :param h: (float) Is a scale variable
        :return: Variables for the matplotlib plotter
        """
        F = (Fx ** 2 + Fz ** 2) ** 0.5
        len_x = Fx / F * h
        len_y = -Fz / F * h
        x = node.vertex.x - len_x * 1.2
        y = node.vertex.y - len_y * 1.2

        return x, y, len_x, len_y, F

    def __point_load_patch(self, max_plot_range, verbosity=0, ax=None):
        """
        :param max_plot_range: max scale of the plot
        """
        if ax is None:
            ax = self.one_fig
        for k in self.system.loads_point:
            Fx, Fz = self.system.loads_point[k]
            F = (Fx ** 2 + Fz ** 2) ** 0.5
            node = self.system.node_map[k]
            h = 0.1 * max_plot_range * F / self.max_system_point_load
            x, y, len_x, len_y, F = self.__arrow_patch_values(Fx, Fz, node, h)

            ax.arrow(
                x,
                y,
                len_x,
                len_y,
                head_width=h * 0.15,
                head_length=0.2 * h,
                ec="b",
                fc="orange",
                zorder=11,
            )
            ax.text(x, y, "F=%d" % F, color="k", fontsize=9, zorder=10)

    def __moment_load_patch(self, max_val, ax=None):
        if ax is None:
            ax = self.one_fig
        h = 0.2 * max_val
        for k, v in self.system.loads_moment.items():

            node = self.system.node_map[k]
            if v > 0:
                ax.plot(
                    node.vertex.x,
                    -node.vertex.z,
                    marker=r"$\circlearrowleft$",
                    ms=25,
                    color="orange",
                )
            else:
                ax.plot(
                    node.vertex.x,
                    -node.vertex.z,
                    marker=r"$\circlearrowright$",
                    ms=25,
                    color="orange",
                )
            ax.text(
                node.vertex.x + h * 0.2,
                -node.vertex.z + h * 0.2,
                "T=%d" % v,
                color="k",
                fontsize=9,
                zorder=10,
            )

    def plot_structure(
        self,
        figsize,
        verbosity,
        show=False,
        supports=True,
        scale=1,
        offset=(0, 0),
        gridplot=False,
        annotations=True,
        ax=None,
        loads=False,
        adjust_ax_width=True,
        adjust_ax_height=True,
    ):
        """
        :param show: (boolean) if True, plt.figure will plot.
        :param supports: (boolean) if True, supports are plotted.
        :param annotations: (boolean) if True, structure annotations are plotted. It includes section name.
        :return:
        """
        ax_is_None = ax is None

        if ax_is_None and not gridplot:
            self.__start_plot(figsize)

        if ax_is_None:
            ax = self.one_fig

        x, y = super().structure()

        max_x = np.max(x)
        min_x = np.min(x)
        max_y = np.max(y)
        min_y = np.min(y)

        center_x = (max_x - min_x) / 2 + min_x + -offset[0]
        center_y = (max_y - min_y) / 2 + min_x + -offset[1]

        max_plot_range = max(max_x, max_y)
        ax_range = max_plot_range * scale

        plusxrange = center_x + ax_range
        plusyrange = center_y + ax_range * figsize[1] / figsize[0]
        minxrange = center_x - ax_range
        minyrange = center_y - ax_range * figsize[1] / figsize[0]

        if adjust_ax_width:
            ax.set(xlim=(minxrange, plusxrange))
        if adjust_ax_height:
            ax.set(ylim=(minyrange, plusyrange))

        for el in self.system.element_map.values():
            x_val, y_val = plot_values_element(el)
            if verbosity == 0:
                ax.plot(x_val, y_val, color="black", marker="s")
                # add node ID to plot
                ax_range = max_plot_range * 0.015
                ax.text(
                    x_val[0] + ax_range,
                    y_val[0] + ax_range,
                    "%d" % el.node_id1,
                    color="g",
                    fontsize=9,
                    zorder=10,
                )
                ax.text(
                    x_val[-1] + ax_range,
                    y_val[-1] + ax_range,
                    "%d" % el.node_id2,
                    color="g",
                    fontsize=9,
                    zorder=10,
                )

                # add element ID to plot
                factor = 0.02 * self.max_val_structure
                x_val = (x_val[0] + x_val[-1]) / 2 - np.sin(el.angle) * factor
                y_val = (y_val[0] + y_val[-1]) / 2 + np.cos(el.angle) * factor

                ax.text(
                    x_val, y_val, str(el.id), color="r", fontsize=9, zorder=10
                )

                # add element annotation to plot
                # TODO: check how this holds with multiple structure scales.
                if annotations:
                    x_val += +np.sin(el.angle) * factor * 2.3
                    y_val += -np.cos(el.angle) * factor * 2.3
                    ax.text(
                        x_val, y_val, el.section_name, color="b", fontsize=9, zorder=10
                    )
            else:
                ax.plot(x_val, y_val, color="black")
        # add supports
        if supports:
            self.__fixed_support_patch(max_plot_range * scale, ax=ax)
            self.__hinged_support_patch(max_plot_range * scale, ax=ax)
            self.__rotational_support_patch(max_plot_range * scale, ax=ax)
            self.__roll_support_patch(max_plot_range * scale, ax=ax)
            self.__rotating_spring_support_patch(max_plot_range * scale, ax=ax)
            self.__spring_support_patch(max_plot_range * scale, ax=ax)

        # add_loads
        if loads or verbosity == 0:
            self.__q_load_patch(max_plot_range, verbosity, ax=ax)
            self.__point_load_patch(max_plot_range, verbosity, ax=ax)
            self.__moment_load_patch(max_plot_range, ax=ax)

        if ax_is_None:
            if show:
                self.plot()
            else:
                return self.fig

    def _add_node_values(self, x_val, y_val, value_1, value_2, digits, ax=None):
        if ax is None:
            ax = self.one_fig
        offset = self.max_val_structure * 0.015

        # add value to plot
        ax.text(
            x_val[1] - offset,
            y_val[1] + offset,
            "%s" % round(value_1, digits),
            fontsize=9,
            ha="center",
            va="center",
        )
        ax.text(
            x_val[-2] - offset,
            y_val[-2] + offset,
            "%s" % round(value_2, digits),
            fontsize=9,
            ha="center",
            va="center",
        )

    def _add_element_values(self, x_val, y_val, value, index, digits=2, ax=None):
        if ax is None:
            ax = self.one_fig

        ax.text(
            x_val[index],
            y_val[index],
            "%s" % round(value, digits),
            fontsize=9,
            ha="center",
            va="center",
        )

    def plot_result(
        self,
        axis_values,
        force_1=None,
        force_2=None,
        digits=2,
        node_results=True,
        fill_polygon=True,
        color=0,
        ax=None,
    ):
        if ax is None:
            ax = self.one_fig
        if fill_polygon:
            rec = plt.Polygon(
                np.vstack(axis_values).T, color="C{}".format(color), alpha=0.3
            )
            ax.add_patch(rec)
        # plot force
        x_val = axis_values[0]
        y_val = axis_values[1]

        ax.plot(x_val, y_val, color="C{}".format(color))

        if node_results:
            self._add_node_values(x_val, y_val, force_1, force_2, digits, ax=ax)

    def plot(self):
        plt.show()

    def axial_force(
        self,
        factor=None,
        figsize=None,
        verbosity=0,
        scale=1,
        offset=(0, 0),
        show=True,
        gridplot=False,
    ):
        self.plot_structure(figsize, 1, scale=scale, offset=offset, gridplot=gridplot)

        if factor is None:
            max_force = max(
                map(
                    lambda el: max(abs(el.N_1), abs(el.N_2)),
                    self.system.element_map.values(),
                )
            )
            factor = det_scaling_factor(max_force, self.max_val_structure)

        for el in self.system.element_map.values():
            if math.isclose(el.N_1, 0, rel_tol=1e-5, abs_tol=1e-9):
                continue
            else:
                axis_values = plot_values_axial_force(el, factor)
                color = 1 if el.N_1 < 0 else 0
                self.plot_result(
                    axis_values,
                    el.N_1,
                    el.N_2,
                    node_results=not bool(verbosity),
                    color=color,
                )

                point = (el.vertex_2 - el.vertex_1) / 2 + el.vertex_1
                if el.N_1 < 0:
                    point.displace_polar(
                        alpha=el.angle + 0.5 * np.pi,
                        radius=0.5 * el.N_1 * factor,
                        inverse_z_axis=True,
                    )

                    if verbosity == 0:
                        self.one_fig.text(
                            point.x,
                            point.y,
                            "-",
                            ha="center",
                            va="center",
                            fontsize=20,
                            color="b",
                        )
                if el.N_1 > 0:
                    point.displace_polar(
                        alpha=el.angle + 0.5 * np.pi,
                        radius=0.5 * el.N_1 * factor,
                        inverse_z_axis=True,
                    )

                    if verbosity == 0:
                        self.one_fig.text(
                            point.x,
                            point.y,
                            "+",
                            ha="center",
                            va="center",
                            fontsize=14,
                            color="b",
                        )

        if show:
            self.plot()
        else:
            return self.fig

    def bending_moment(
        self,
        factor=None,
        figsize=None,
        verbosity=0,
        scale=1,
        offset=(0, 0),
        show=True,
        gridplot=False,
        ax=None,
        adjust_ax_height=True,
    ):
        self.plot_structure(figsize, 1, supports=False, adjust_ax_height=adjust_ax_height,
                            scale=scale, offset=offset, gridplot=gridplot, ax=ax)

        ax_is_None = ax is None
        if ax_is_None:
            ax = self.one_fig

        con = len(self.system.element_map[1].bending_moment)
        if factor is None:
            # maximum moment determined by comparing the node's moments and the sagging moments.
            max_moment = max(
                map(
                    lambda el: max(
                        abs(el.node_1.Ty),
                        abs(((el.all_q_load[0] + el.all_q_load[1]) / 16) * el.l ** 2),
                    )
                    if el.type == "general"
                    else 0,
                    self.system.element_map.values(),
                )
            )
            factor = det_scaling_factor(max_moment, self.max_val_structure)

        # determine the axis values
        for el in self.system.element_map.values():
            if (
                math.isclose(el.node_1.Ty, 0, rel_tol=1e-5, abs_tol=1e-9)
                and math.isclose(el.node_2.Ty, 0, rel_tol=1e-5, abs_tol=1e-9)
                and not el.all_q_load[0]
                and not el.all_q_load[1]
            ):
                # If True there is no bending moment, so no need for plotting.
                continue
            axis_values = plot_values_bending_moment(el, factor, con)
            if verbosity == 0:
                node_results = True
            else:
                node_results = False
            self.plot_result(
                axis_values,
                abs(el.node_1.Ty),
                abs(el.node_2.Ty),
                node_results=node_results,
                ax=ax
            )

            if el.all_q_load:
                m_sag = min(el.bending_moment)
                index = find_nearest(el.bending_moment, m_sag)[1]
                offset = -self.max_val_structure * 0.05

                if verbosity == 0:
                    x = axis_values[0][index] + np.sin(-el.angle) * offset
                    y = axis_values[1][index] + np.cos(-el.angle) * offset
                    ax.text(x, y, "%s" % round(m_sag, 1), fontsize=9)

        # Apply equal padding up and below the plot
        self._add_vertical_padding(ax)

        if ax_is_None:
            if show:
                self.plot()
            else:
                return self.fig

    def _add_vertical_padding(self, ax):
        ymin, ymax = ax.get_ylim()
        max_diff = 0.5 * max(abs(ymin), abs(ymax))
        ax.set(ylim=(ymin + np.sign(ymin) * max_diff, ymax + ymax + np.sign(ymax) * max_diff))

    def shear_force(
        self,
        factor=None,
        figsize=None,
        verbosity=0,
        scale=1,
        offset=(0, 0),
        show=True,
        gridplot=False,
        include_structure=True,
        adjust_ax_height=True,
        ax=None,
    ):
        ax_is_None = ax is None
        if ax_is_None:
            ax = self.one_fig

        if include_structure:
            self.plot_structure(
                figsize, 1, scale=scale, offset=offset, gridplot=gridplot, ax=ax,
                adjust_ax_height=adjust_ax_height, supports=False
            )
        if factor is None:
            max_force = max(
                map(
                    lambda el: np.max(np.abs(el.shear_force)),
                    self.system.element_map.values(),
                )
            )
            factor = det_scaling_factor(max_force, self.max_val_structure)

        for el in self.system.element_map.values():
            if (
                math.isclose(el.node_1.Ty, 0, rel_tol=1e-5, abs_tol=1e-9)
                and math.isclose(el.node_2.Ty, 0, rel_tol=1e-5, abs_tol=1e-9)
                and el.q_load is None
            ):
                # If True there is no bending moment and no shear, thus no shear force, so no need for plotting.
                continue
            axis_values = plot_values_shear_force(el, factor)
            shear_1 = el.shear_force[0]
            shear_2 = el.shear_force[-1]

            self.plot_result(
                axis_values, shear_1, shear_2, node_results=not bool(verbosity), ax=ax
            )

        # Apply equal padding up and below the plot
        self._add_vertical_padding(ax)

        if ax_is_None:
            if show:
                self.plot()
            else:
                return self.fig

    def reaction_force(self, figsize, verbosity, scale, offset, show, gridplot=False):
        self.plot_structure(
            figsize, 1, supports=False, scale=scale, offset=offset, gridplot=gridplot
        )

        h = 0.2 * self.max_val_structure
        max_force = max(
            map(
                lambda node: max(abs(node.Fx), abs(node.Fz)),
                self.system.reaction_forces.values(),
            )
        )

        for node in self.system.reaction_forces.values():
            if not math.isclose(node.Fx, 0, rel_tol=1e-5, abs_tol=1e-9):
                # x direction
                scale = abs(node.Fx) / max_force * h
                sol = self.__arrow_patch_values(node.Fx, 0, node, scale)
                x = sol[0]
                y = sol[1]
                len_x = sol[2]
                len_y = sol[3]

                self.one_fig.arrow(
                    x,
                    y,
                    len_x,
                    len_y,
                    head_width=h * 0.15,
                    head_length=0.2 * scale,
                    ec="b",
                    fc="orange",
                    zorder=11,
                )

                if verbosity == 0:
                    self.one_fig.text(
                        x,
                        y,
                        "R=%s" % round(node.Fx, 2),
                        color="k",
                        fontsize=9,
                        zorder=10,
                    )

            if not math.isclose(node.Fz, 0, rel_tol=1e-5, abs_tol=1e-9):
                # z direction
                scale = abs(node.Fz) / max_force * h
                sol = self.__arrow_patch_values(0, node.Fz, node, scale)
                x = sol[0]
                y = sol[1]
                len_x = sol[2]
                len_y = sol[3]

                self.one_fig.arrow(
                    x,
                    y,
                    len_x,
                    len_y,
                    head_width=h * 0.15,
                    head_length=0.2 * scale,
                    ec="b",
                    fc="orange",
                    zorder=11,
                )

                if verbosity == 0:
                    self.one_fig.text(
                        x,
                        y,
                        "R=%s" % round(node.Fz, 2),
                        color="k",
                        fontsize=9,
                        zorder=10,
                    )

            if not math.isclose(node.Ty, 0, rel_tol=1e-5, abs_tol=1e-9):
                """
                '$...$': render the strings using mathtext
                """
                if node.Ty > 0:
                    self.one_fig.plot(
                        node.vertex.x,
                        -node.vertex.z,
                        marker=r"$\circlearrowleft$",
                        ms=25,
                        color="orange",
                    )
                if node.Ty < 0:
                    self.one_fig.plot(
                        node.vertex.x,
                        -node.vertex.z,
                        marker=r"$\circlearrowright$",
                        ms=25,
                        color="orange",
                    )

                if verbosity == 0:
                    self.one_fig.text(
                        node.vertex.x + h * 0.2,
                        -node.vertex.z + h * 0.2,
                        "T=%s" % round(node.Ty, 2),
                        color="k",
                        fontsize=9,
                        zorder=10,
                    )
        if show:
            self.plot()
        else:
            return self.fig

    def displacements(
        self,
        factor=None,
        figsize=None,
        verbosity=0,
        scale=1,
        offset=(0, 0),
        show=True,
        linear=False,
        gridplot=False,
    ):
        self.plot_structure(figsize, 1, scale=scale, offset=offset, gridplot=gridplot)
        if factor is None:
            # needed to determine the scaling factor
            max_displacement = max(
                map(
                    lambda el: max(
                        abs(el.node_1.ux), abs(el.node_1.uz), el.max_deflection
                    )
                    if el.type == "general"
                    else 0,
                    self.system.element_map.values(),
                )
            )
            factor = det_scaling_factor(max_displacement, self.max_val_structure)

        for el in self.system.element_map.values():
            axis_values = plot_values_deflection(el, factor, linear)
            self.plot_result(axis_values, node_results=False, fill_polygon=False)

            if el.type == "general":
                # index of the max deflection
                x = np.linspace(el.vertex_1.x, el.vertex_2.x, el.deflection.size)
                y = np.linspace(el.vertex_1.y, el.vertex_2.y, el.deflection.size)
                xd, yd = plot_values_deflection(el, 1.0, linear)
                deflection = ((xd - x) ** 2 + (yd - y) ** 2) ** 0.5
                index = np.argmax(np.abs(deflection))

                if verbosity == 0:

                    if index != 0 or index != el.deflection.size:
                        self._add_element_values(
                            axis_values[0],
                            axis_values[1],
                            deflection[index],
                            index,
                            3,
                        )
        if show:
            self.plot()
        else:
            return self.fig

    def results_plot(self, figsize, verbosity, scale, offset, show):
        """
        Aggregate all the plots in one grid plot.

        :param figsize: (tpl)
        :param verbosity: (int)
        :param scale: (flt)
        :param offset: (tpl)
        :param show: (bool)
        :return: Figure or None
        """
        plt.close("all")
        self.fig = plt.figure(figsize=figsize)
        a = 320
        self.one_fig = self.fig.add_subplot(a + 1)
        plt.title("structure")
        self.plot_structure(
            figsize, verbosity, show=False, scale=scale, offset=offset, gridplot=True
        )
        self.one_fig = self.fig.add_subplot(a + 2)
        plt.title("bending moment")
        self.bending_moment(None, figsize, verbosity, scale, offset, False, True)
        self.one_fig = self.fig.add_subplot(a + 3)
        plt.title("shear force")
        self.shear_force(None, figsize, verbosity, scale, offset, False, True)
        self.one_fig = self.fig.add_subplot(a + 4)
        plt.title("axial force")
        self.axial_force(None, figsize, verbosity, scale, offset, False, True)
        self.one_fig = self.fig.add_subplot(a + 5)
        plt.title("displacements")
        self.displacements(None, figsize, verbosity, scale, offset, False, False, True)
        self.one_fig = self.fig.add_subplot(a + 6)
        plt.title("reaction force")
        self.reaction_force(figsize, verbosity, scale, offset, False, True)

        if show:
            self.plot()
        else:
            return self.fig


import numpy as np
import traits.api as tr
from anastruct import SystemElements
from bmcs_beam.beam_config.system.system import System
from bmcs_utils.api import Int, Item, View, Float, FloatEditor, IntEditor


class BeamSystem(System):

    name = 'BeamSystem'

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.update_struct()

    L = Float(6000, SYSTEM=True)
    n_x = Int(10, desc='Number of FE elements along the beam.', SYSTEM=True)
    struct = SystemElements()
    _n_x_min = 5

    tree = []

    ipw_view = View(
        Item('L', latex='L \mathrm{[mm]}', editor=FloatEditor(min=10)),
        # Item('n_x', latex='n_x', editor=IntEditor(min_name='_n_x_min')),
        *System.ipw_view.content,
    )

    x = tr.Property(depends_on='L, n_x, n_per_element')
    @tr.cached_property
    def _get_x(self):
        return np.linspace(0, self.L, self.n_x * self.n_per_element)

    def _get_L(self):
        return (self.struct.element_map[self.n_x].vertex_2.x - self.struct.element_map[1].vertex_1.x)

    def subplots(self, fig):
        axes = fig.subplots(3, 1)
        return axes

    def update_plot(self, axes):
        self.plot_struct(axes[0])
        self.plot_bending_moment(axes[1])
        self.plot_shear(axes[2])

    def plot_struct(self, ax):
        ax.get_yaxis().set_visible(False)
        self.struct.plotter.plot_structure(figsize=(8, 2), verbosity=1, loads=True, ax=ax)
        ax.set_xlabel('x [mm]')

    def plot_shear(self, ax):
        # factor=1 is mandatory, here factor=1/1000 is applied [N->kN]
        self.struct.plotter.shear_force(figsize=(8, 2), factor=1/1000, adjust_ax_height=False,
                                           verbosity=1, ax=ax, show=False)
        ax.set_ylabel('Q [kN]')

    def plot_bending_moment(self, ax):
        # Note: factor with (-1) was used to show moment according to our convention
        # factor=1 is mandatory, here factor=1/1e6 is applied [Nmm->kNm]
        self.struct.plotter.bending_moment(figsize=(8, 2), factor=-1/1e6, adjust_ax_height=False,
                                           verbosity=1, ax=ax, show=False)
        ax.set_ylabel('M [kNm]')
        ax.invert_yaxis()
        # M_x = np.array(self.struct.get_element_result_range(unit = "moment")) / M_scale

    # def get_moment(self, n=100):
    #     m = np.array([dict['M'] for dict in self.struct.get_element_results(verbose=True)]).flatten()
    #     x = np.linspace(0, self.L, m.size)
    #     return np.interp(np.linspace(0, self.L, n), x, m)

    def get_moment(self):
        # Note: factor=-1 was used to get the moment sign according to our convention
        return - np.array([dict['M'] for dict in self.struct.get_element_results(verbose=True)]).flatten()

    def get_shear(self):
        return np.array([dict['Q'] for dict in self.struct.get_element_results(verbose=True)]).flatten()

    def get_max_force(self, M_I):
        """ Where M_I the array of moment values corresponding to the studies curvature range [kappa_min, kappa_max] """
        return NotImplementedError
        # elif self.beam_design.beam_conf_name == BoundaryConfig.THREE_SPAN_DIST_LOAD:
        #     # maximum negative moment M_I[0]
        #     F_max = 10 * M_I[0] / self.beam_design.L**2  # max moment is in the 2nd support
        # elif self.beam_design.beam_conf_name == BoundaryConfig.FIXED_SUPPORT_DIST_LOAD:
        #     F_max = 24 * M_I[-1] / self.beam_design.L**2 # max moment is in the span middle
        # elif self.beam_design.beam_conf_name == BoundaryConfig.FIXED_AND_ROLLER_SUPPORT_DIST_LOAD:
        #     # maximum negative moment M_I[0]
        #     F_max = 8 * M_I[0] / self.beam_design.L**2 # max moment is in fixed support

    def get_plot_force_scale_and_unit(self):
        """ Scale which should be applied on the force when plotting """
        return NotImplementedError